package ac.ed.lurg.country;

import java.util.HashMap;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.demand.AbstractDemandManager;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;

public abstract class AbstractCountryAgent {

	protected AbstractDemandManager demandManager;
	protected CompositeCountry country;
	protected Map<CropType, Double> tradeBarriers;
	protected Map<CommodityType, Double> currentProjectedDemand;
	protected Map<CropType, CountryPrice> currentCountryPrices;
	protected Map<CommodityType, Map<CropType, Double>> baseDemandFact;
	protected Timestep currentTimestep;
	protected Map<CommodityType, Map<CropType, Double>> currentMinDemandFract;
	
	public AbstractCountryAgent(AbstractDemandManager demandManager,CompositeCountry country, Map<CropType, Double> tradeBarriers) {

		this.demandManager = demandManager;
		this.country = country;
		this.tradeBarriers = tradeBarriers;
		baseDemandFact = demandManager.getBaseDemandFracts(country);
	}

	public CompositeCountry getCountry() {
		return country;
	}
	
	protected void setCurrentTimestep(Timestep timestep) {
		currentTimestep = timestep;
		currentMinDemandFract = getMinDemandFraction();
	}

	private void calculateCountryPrices(Map<CropType, GlobalPrice> worldPrices){
		Map<CropType, CountryPrice> countryPrices = new HashMap <CropType, CountryPrice>();

		for (CropType c : CropType.getImportedTypes()) {
			GlobalPrice worldPrice = worldPrices.get(c);
			Double tb = tradeBarriers.get(c);
			LogWriter.println(worldPrice + " " + c);
			CountryPrice prices = new CountryPrice(worldPrice.getCountryImportPrice(tb==null ? 0 : tb, currentTimestep), worldPrice.getExportPrice());
			countryPrices.put(c, prices);
		}

		currentCountryPrices = countryPrices;
	}
	
	private void calculateDemand() {
		Map<CommodityType, Double> producerPrices = getProducerCommodityPrices();
		currentProjectedDemand = demandManager.getDemand(country, currentTimestep.getYear(), producerPrices);
	}
	
	protected void calculateCountryPricesAndDemand(Map<CropType, GlobalPrice> worldPrices) {
		calculateCountryPrices(worldPrices);
		calculateDemand();
	}
	
	private Map<CommodityType, Double> getProducerCommodityPrices() {
		if (!ModelConfig.PRICE_ELASTIC_DEMAND)
			return null;
						
		Map<CommodityType, Double> prices = new HashMap<CommodityType, Double>();
		
		for (CommodityType commodity : CommodityType.getAllFoodItems()) {
			double commPricePlum = getCommPriceFromCropPrice(commodity);
			prices.put(commodity, commPricePlum);
			LogWriter.println("Producer price for " + commodity.getGamsName() + " is " + commPricePlum);
		}
		
		return prices;
	}
	
	protected abstract double getCommPriceFromCropPrice(CommodityType commodity);

	public Map<CommodityType, Double> getCurrentProjectedDemand() {
		return currentProjectedDemand;
	}
	
	public Map<CropType, CountryPrice> getCurrentCountryPrices() {
		return currentCountryPrices;
	}
	
	private Map<CommodityType, Map<CropType, Double>> getMinDemandFraction() {
		Map<CommodityType, Map<CropType, Double>> minDemandFraction = new HashMap<CommodityType, Map<CropType, Double>>();

		int yearsOfChange = ModelConfig.IS_CALIBRATION_RUN ? 0
				: (currentTimestep.getTimestep() - ModelConfig.DEMAND_FRACT_CHANGE_START_STEP) * ModelConfig.TIMESTEP_SIZE;

		for (CommodityType comm : baseDemandFact.keySet()) {
			if (comm == CommodityType.CEREALS || comm == CommodityType.OILCROPSPULSES) {
				Map<CropType, Double> cropFracts = new HashMap<CropType, Double>();
				for (Map.Entry<CropType, Double> crop : baseDemandFact.get(comm).entrySet()) {
					cropFracts.put(crop.getKey(), crop.getValue() * Math.max(ModelConfig.TOTAL_MAX_DEMAND_FRACT_CHANGE,
							(1 - ModelConfig.ANNUAL_MAX_DEMAND_FRACT_CHANGE * yearsOfChange)));
				}
				minDemandFraction.put(comm, cropFracts);
			}
		}

		return minDemandFraction;
	}

	protected void updateNetImportsFromProdAndDemand(Map<CommodityType, Double> demands, Map<CommodityType, Map<CropType, Double>>  minDemandFracts, Map<CropType, CropUsageData> cropUsages) {

		for (CommodityType commodity : CommodityType.getAllFoodItems()) {
				
			if (commodity != CommodityType.CEREALS && commodity != CommodityType.OILCROPSPULSES && commodity.getCropTypes().size() != 1)
				throw new RuntimeException("Not cereal or oilcropspulses and not 1 to 1 mapping for commodity to crop: " + commodity);  // skips cereals which is a special case and handled separately below

			double demand = demands.get(commodity);
			if (commodity == CommodityType.CEREALS || commodity == CommodityType.OILCROPSPULSES) {
				Map<CropType, Double> netImportsFromMinDemands = new HashMap<CropType, Double>();
				double totalProd = 0, totalmportFromMD = 0, totalExcessProd = 0;

				for (CropType crop : commodity.getCropTypes()) {
					CropUsageData cropUsage = cropUsages.get(crop);
					double prod = cropUsage.getProductionExpected() *(1-crop.getSeedAndWasteRate()) - cropUsage.getMonogastricFeed() - cropUsage.getRuminantFeed();
					totalProd += prod;
					double minFract = minDemandFracts.get(commodity).containsKey(crop) ? minDemandFracts.get(commodity).get(crop) : 0.0;
					double netImportsFromMinDemand = minFract * demand - prod;
					netImportsFromMinDemands.put(crop, netImportsFromMinDemand);
					
					if (netImportsFromMinDemand > 0)
						totalmportFromMD += netImportsFromMinDemand;
					else
						totalExcessProd += netImportsFromMinDemand;
				}
				double additionalNetImportsRequired = demand - totalProd - totalmportFromMD;
				LogWriter.println("additionalNetImportsRequired is " + additionalNetImportsRequired);

				for (CropType crop : commodity.getCropTypes()) {
					double netImportsMD = netImportsFromMinDemands.get(crop);
					double netImports=0;
					if (additionalNetImportsRequired > 0) { // overall need to import more
						if (minDemandFracts.get(commodity).containsKey(crop))
							netImports = netImportsMD + minDemandFracts.get(commodity).get(crop) * additionalNetImportsRequired; // divide required additional imports by minCerealFracts		
					}
					else {  // overall need to export more
						if (netImportsMD > 0)
							netImports = netImportsMD; // still import what we need for minimum fraction
						else if(netImportsMD < 0)
							netImports = netImportsMD / totalExcessProd * additionalNetImportsRequired; // divide exports by production in excess of minimum demand
					}
					
					CropUsageData cropUsage = cropUsages.get(crop);	
					LogWriter.println("Updating " + commodity + " net imports " + crop + " to " + netImports);
					cropUsage.updateNetImports(netImports);
				}
			}
			else {
				// simple 1-1 commodity to cereal mappings
				for (CropType crop : commodity.getCropTypes()) {
					CropUsageData cropUsage = cropUsages.get(crop);
					double prod =  cropUsage.getProductionExpected() *(1-crop.getSeedAndWasteRate()) - cropUsage.getMonogastricFeed() - cropUsage.getRuminantFeed();
					double netImports = demand - prod;
					LogWriter.println("Updating net imports " + crop + " to " + netImports);
					cropUsage.updateNetImports(netImports);
				}
			} 
		} 
	}
	
	abstract public Map<CropType, CropUsageData> getCropUsageData();
}