package ac.ed.lurg.country;

import java.io.File;
import java.io.IOException;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.gams.GamsCountryInput;
import ac.ed.lurg.country.gams.GamsRasterInput;
import ac.ed.lurg.country.gams.GamsRasterOptimiser;
import ac.ed.lurg.country.gams.GamsRasterOutput;
import ac.ed.lurg.demand.AbstractDemandManager;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.landuse.IrrigationItem;
import ac.ed.lurg.landuse.LandUseItem;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.cluster.Cluster;
import ac.ed.lurg.utils.cluster.KMeans;
import ac.ed.lurg.yield.YieldClusterPoint;
import ac.ed.lurg.yield.YieldRaster;
import ac.ed.lurg.yield.YieldResponsesItem;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class CountryAgent extends AbstractCountryAgent {

	private GamsRasterOutput previousGamsRasterOutput;
	private RasterSet<IntegerRasterItem> yieldClusters;
	private double animalFeedFromOtherSources;
	private Map<CropType, Double> subsidyRates;
	private boolean saveGamsGdxFiles;

	public CountryAgent(AbstractDemandManager demandManager,CompositeCountry country, RasterSet<LandUseItem> cropAreaRaster,
			Map<CropType, CropUsageData> cropUsageData, Map<CropType, Double> tradeBarriers, RasterSet<IntegerRasterItem> yieldClusters,
			Map<CropType, Double> subsidyRates) {

		super(demandManager, country, tradeBarriers);
		this.yieldClusters = yieldClusters;
		this.subsidyRates = subsidyRates;

		GamsRasterOutput initialData = new GamsRasterOutput(cropAreaRaster, cropUsageData);
		previousGamsRasterOutput = initialData;
		
		saveGamsGdxFiles = (ModelConfig.GAMS_COUNTRY_TO_SAVE != null && country.getName().equals(ModelConfig.GAMS_COUNTRY_TO_SAVE));
	}

	public RasterSet<IntegerRasterItem> getYieldClusters() {
		return yieldClusters;
	}

	private RasterSet<IntegerRasterItem> calcYieldClusters(RasterSet<IrrigationItem> irrigData, YieldRaster countryYieldSurfaces) {

		LogWriter.println("calcYieldClusters: for " + ModelConfig.NUM_YIELD_CLUSTERS + " clusters");	

		// create collection of ClusteringPoints from countryYieldSurfaces, these have the RasterKey and data for yield (or access to them)
		Collection<YieldClusterPoint> clusteringPoints = new HashSet<YieldClusterPoint>();
		for (Entry<RasterKey, YieldResponsesItem> entry : countryYieldSurfaces.entrySet()) {
			YieldResponsesItem yieldresp = entry.getValue();
			if (yieldresp != null) {
				RasterKey key = entry.getKey();
				IrrigationItem irrigItem = irrigData.get(key);
				
				LandUseItem luItem = previousGamsRasterOutput.getLandUses().get(key);
				double protectedArea = luItem.getProtectedArea();
				clusteringPoints.add(new YieldClusterPoint(key, yieldresp, irrigItem, protectedArea));
			}
		}

		// do the clustering
		KMeans<String, YieldClusterPoint> kmeans = new KMeans<String, YieldClusterPoint>(clusteringPoints, ModelConfig.NUM_YIELD_CLUSTERS);
		kmeans.calculateClusters(100, 0.1);
		kmeans.printClusters();

		// reformat output
		List<Cluster<String, YieldClusterPoint>> yieldClusters = kmeans.getClusters();
		RasterSet<IntegerRasterItem> mapping = new RasterSet<IntegerRasterItem>(countryYieldSurfaces.getHeaderDetails());

		int id = 1;
		for (Cluster<String, YieldClusterPoint> c : yieldClusters) {
			for (YieldClusterPoint p : c.getPoints())
				mapping.put(p.getRasterKey(), new IntegerRasterItem(id));

			if (c.getPoints().size()>0)
				id++;
		}

		return mapping;
	}

	public GamsRasterOutput determineProduction(YieldRaster countryYieldSurfaces, RasterSet<IrrigationItem> irrigData, 
			Map<CropType, GlobalPrice> worldPrices, double globalGen2EcIncrease) {

		// project demand
		calculateCountryPricesAndDemand(worldPrices);
		
		if (saveGamsGdxFiles && ModelConfig.PRICE_ELASTIC_DEMAND)
			saveGDXFile("demand");
		
		if (currentTimestep.isInitialTimestep()) {
			double totalAnimalProductDemand = currentProjectedDemand.get(CommodityType.MONOGASTRICS) + currentProjectedDemand.get(CommodityType.RUMINANTS);
			animalFeedFromOtherSources = totalAnimalProductDemand * ModelConfig.ANIMAL_FEED_FROM_OTHER_SOURCES_RATE;
		}
						
		if (currentProjectedDemand.size() == 0) {
			LogWriter.printlnError("No demand for country " + country + " so skipping it");
		}
		else if (countryYieldSurfaces.size() == 0 ) {
			LogWriter.printlnError("No yield values for " + country + " so skipping it");
		}
		else if (ModelConfig.DEBUG_JUST_DEMAND_OUTPUT) { // if this debug flag is set we don't do the optimisation
		}
		else {			
			if (yieldClusters==null)
				yieldClusters = calcYieldClusters(irrigData, countryYieldSurfaces);  // this should only be on the first timestep

			// optimize areas and intensity 
			GamsRasterInput input = getGamsRasterInput(irrigData, countryYieldSurfaces, globalGen2EcIncrease);
			GamsRasterOptimiser opti = new GamsRasterOptimiser(input, yieldClusters);
			LogWriter.println("Running " + country.getName() + ", currentTimestep " + currentTimestep);

			GamsRasterOutput result = opti.run();

			if (saveGamsGdxFiles)
				saveGDXFile("landuse");
			
			previousGamsRasterOutput = result;
			return result;
		}

		throw new RuntimeException("Skipping optimisation of country " + country);
	}
	
	public void updateNetImportsFromProdAndDemand() {
		updateNetImportsFromProdAndDemand(currentProjectedDemand, currentMinDemandFract, previousGamsRasterOutput.getCropUsageData());
	}
	
	private void saveGDXFile(String ext) {
		// some hacky code for debug purposes that keeps each gams gdx file
		try {
			Files.copy(
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + "_gams_java_gdb1.gdx"),
					FileSystems.getDefault().getPath(ModelConfig.TEMP_DIR + File.separator + country.getName().replaceAll("\\s+","") + currentTimestep.getYear() + ext + ".gdx"),
					StandardCopyOption.REPLACE_EXISTING);
		} catch (IOException e) {
			LogWriter.print(e);
		}
	}

	private GamsRasterInput getGamsRasterInput(RasterSet<IrrigationItem> irrigData, YieldRaster countryYieldSurfaces, double gen2EcIncrease) {
		double allowedImportChange;

		if (currentTimestep.isInitialTimestep()) {  // initialisation time-step
			allowedImportChange = 0.0;
		}
		else { // normal (not the initial) time-step
			allowedImportChange = ModelConfig.MAX_IMPORT_CHANGE;  // when running is calibration model calibrate (ModelConfig.IS_CALIBRATION_RUN==true) MAX_IMPORT_CHANGE is already 0.
		}

		Map<CropType, TradeConstraint> importConstraints = new HashMap<CropType, TradeConstraint>();

		for (Map.Entry<CropType, CropUsageData> entry : previousGamsRasterOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			CropType crop = entry.getKey();
			double baseTrade = cropUsage.getNetImportsExpected();
			double changeUp = 0.0;
			double changeDown = 0.0;

			if (allowedImportChange > 0.0) {
				double maxOfProdOrSupply = cropUsage.getProductionExpected() + Math.max(baseTrade, 0);  //max of supply for food

				// max of supply overall, needed for when imports are supplying feed and change is zero to allow for production to replace imports
		// CONFUSING NOT SURE IF NEEDED		if (maxOfProdOrSupply == 0) maxOfProdOrSupply = Math.max(cropUsage.getNetImports() + cropUsage.getProduction() , cropUsage.getProduction());  
				
				changeDown = changeUp = allowedImportChange * maxOfProdOrSupply;
			}
			if (CropType.ENERGY_CROPS.equals(crop)) {
				double croplandArea = LandUseItem.getTotalLandCover(previousGamsRasterOutput.getLandUses().values(), LandCoverType.CROPLAND);
				double ecMaxExport = gen2EcIncrease * croplandArea / 1500.0 * 2;
				if (ModelConfig.IS_CALIBRATION_RUN) {
					baseTrade = 0.0;
					ecMaxExport = 0.0; 
				}
				changeDown = ecMaxExport;

				if (Double.isNaN(baseTrade))
					baseTrade=0;
			}

			importConstraints.put(crop, new TradeConstraint(baseTrade - changeDown, baseTrade + changeUp));
		}

		GamsCountryInput countryLevelInputs = new GamsCountryInput(country, currentProjectedDemand, currentCountryPrices, importConstraints, 
				previousGamsRasterOutput.getCropUsageData(), currentMinDemandFract, animalFeedFromOtherSources, subsidyRates);	
		GamsRasterInput input = new GamsRasterInput(currentTimestep, countryYieldSurfaces, previousGamsRasterOutput.getLandUses(), irrigData, countryLevelInputs);

		return input;
	}
	
	@Override
	protected double getCommPriceFromCropPrice(CommodityType commodity) {
		double commPricePlum = 0;
		Map<CropType, Double> demandFract = baseDemandFact.get(commodity);
		Map<CropType, CropUsageData> cropUsageMap = previousGamsRasterOutput.getCropUsageData();
		
		for (CropType crop : commodity.getCropTypes()) {
			CropUsageData cropUsageData = cropUsageMap.get(crop);
	
			double weighting = Math.max(0, cropUsageData.getShockedNetImports()/(cropUsageData.getProductionExpected()+cropUsageData.getNetImportsExpected()));
			double newImportPrice = currentCountryPrices.get(crop).getImportPrice()*weighting + currentCountryPrices.get(crop).getExportPrice()*(1-weighting);
			
			commPricePlum += newImportPrice * demandFract.get(crop);  // weight price by base demand of each cereal crop
		}
		
		return commPricePlum;
	}
	
	public void shockGDP() {

		double totalProduction = getTotalProduction();
		double shockMagnitude = 0.0;
		Map<CropType, CropUsageData> cropUsages = previousGamsRasterOutput.getCropUsageData();

		for (CropType crop : CropType.getImportedTypes())  {	

			if(cropUsages.get(crop).getProductionExpected() != 0)
				shockMagnitude += (cropUsages.get(crop).getProductionShock() /cropUsages.get(crop).getProductionExpected()) * (cropUsages.get(crop).getProductionExpected()/totalProduction);	
		}	
		demandManager.updateGdpLossesFromShock(country, currentTimestep.getYear(), shockMagnitude);
	}

	public RasterSet<LandUseItem> getLandUses() {
		return previousGamsRasterOutput.getLandUses();
	}

	public Map<CropType, CropUsageData> getCropUsageData() {
		return previousGamsRasterOutput.getCropUsageData();
	}
	
	public double getTotalProduction() {
		double totalProduction = 0;
		for (CropType crop : CropType.getImportedTypes()) { //assuming pasture and set aside not part of agricultural value, valid or not?
			totalProduction +=  previousGamsRasterOutput.getCropUsageData().get(crop).getProductionExpected();
		}	
		return totalProduction;
	}
}