package ac.ed.lurg.country;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.types.Parameter;
import ac.ed.lurg.utils.LogWriter;

public class GlobalPrice {
	private double exportPrice;
	private double importAmount;
	private double exportAmountBeforeLoss;
	private double transportLosses;
	private double stockLevel;

	private GlobalPrice(double exportPrice, double stockLevel, double importAmount, double exportAmountBeforeLoss, double transportLosses) {
		this.exportPrice = exportPrice;
		this.stockLevel = stockLevel;
		this.importAmount = importAmount;
		this.exportAmountBeforeLoss = exportAmountBeforeLoss;
		this.transportLosses = transportLosses;
	}
	
	public static GlobalPrice createInitial(double exportPrice) {
		return new GlobalPrice(exportPrice, Double.NaN, Double.NaN, Double.NaN, Double.NaN);
	}
	
	public double getExportPrice(){
		return exportPrice;
	}

	public double getStockLevel(){
		return stockLevel;
	}
	
	public double getCountryImportPrice(double countryTradeBarrier, Timestep timestep) {

		int currentYear = timestep.getYear();
		
		double tradeBarrierMultiplier = (!ModelConfig.SHOCKS_POSSIBLE)? ModelConfig.TRADE_BARRIER_MULTIPLIER: ModelConfig.getParameter(Parameter.TRADE_BARRIER_MULTIPLIER, currentYear);
		double transportLossesRate =(!ModelConfig.SHOCKS_POSSIBLE) ? ModelConfig.TRANSPORT_LOSSES : ModelConfig.getParameter(Parameter.TRANSPORT_LOSSES, currentYear);
		double transportCosts = (!ModelConfig.SHOCKS_POSSIBLE) ? ModelConfig.TRANSPORT_COST : ModelConfig.getParameter(Parameter.TRANSPORT_COST, currentYear);
		
		double importPrice = exportPrice * (1 + transportCosts) * (1.0 + countryTradeBarrier*tradeBarrierMultiplier) 
								/ (1.0 - transportLossesRate);
		return importPrice;
	}

	public double getImportAmount() {
		return importAmount;
	}
	
	public double getExportsAfterTransportLosses() {
		return exportAmountBeforeLoss - transportLosses;
	}
	
	public double getExportsBeforeTransportLoss() {
		return exportAmountBeforeLoss ;
	}
	
	public double getTransportLosses() {
		return transportLosses;
	}

	public GlobalPrice createWithUpdatedMarketPrices(double imports, double exportsBeforeTransportLosses, Timestep timestep, double production) {

		if (Double.isNaN(stockLevel))
			stockLevel = ModelConfig.STOCK_TARGET_RATE * production;  // set stock initially to target level, this is pretty horrible and should really be done on construction but don't easily have production value at that point

		double transportLossRate = (!ModelConfig.SHOCKS_POSSIBLE) ? ModelConfig.TRANSPORT_LOSSES : ModelConfig.getParameter(Parameter.TRANSPORT_LOSSES, timestep.getYear());
		double transportLosses = exportsBeforeTransportLosses * transportLossRate;
		double exportsAfterTransportLosses = exportsBeforeTransportLosses - transportLosses;
		
		if (imports > 0 || exportsAfterTransportLosses > 0) {
			double updatedStock = stockLevel + exportsBeforeTransportLosses - transportLosses - imports;
			LogWriter.println(String.format(" global stocks from %.2f to %.2f", stockLevel, updatedStock));
			if (updatedStock < 0)
				LogWriter.println(String.format("Global stocks are below zero"));
			
			double newPrice, adjustment;
			double financialSpeculationMultiplier = (!ModelConfig.SHOCKS_POSSIBLE)? 1.0 : ModelConfig.getParameter(Parameter.FINANCIAL_SPECULATION_MULTIPLIER,timestep.getYear());

			if (!ModelConfig.MARKET_ADJ_PRICE) {
				adjustment = 1;
			}
			else if (ModelConfig.PRICE_UPDATE_BY_MARKET_IMBALANCE) {
				double ratio = (imports-exportsAfterTransportLosses)/(imports-exportsAfterTransportLosses+production);
				adjustment =  Math.exp(ratio * ModelConfig.MARKET_LAMBA);
				LogWriter.println(String.format("Price update market imbalance: adjustment= %s, imports %.1f, exports %.1f", adjustment, imports, getExportsAfterTransportLosses()));
			}
			else {
				double stockTarget = ModelConfig.STOCK_TARGET_RATE * production;
				double ratio = (stockTarget - updatedStock) / stockTarget;
				
				if (Math.abs(ratio) >= 1)
					adjustment = 10;  // default to max change
				else
					adjustment = 1 + ModelConfig.MARKET_LAMBA * ratio/(1 - ratio*ratio);
				
				LogWriter.println(String.format("Price update from stock target: adjustment= %s, ratio= %s", adjustment, ratio));
			}
			
			newPrice = exportPrice * adjustment * financialSpeculationMultiplier;
			return new GlobalPrice(newPrice, updatedStock, imports, exportsBeforeTransportLosses, transportLosses);			
		}
		else {
			LogWriter.printlnError(String.format("Price for not updated (still %s), as no imports and no exports for it", exportPrice));
			return this;
		}
	}

	@Override
	public String toString() {
		return String.format("exportPrice=%.6f", exportPrice);
	}

	public double getStockChange() {
		return exportAmountBeforeLoss - transportLosses - importAmount;
	}
}
