package ac.ed.lurg.demand;

import java.util.HashMap;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.country.CompositeCountry;
import ac.ed.lurg.country.CompositeCountryManager;
import ac.ed.lurg.country.SingleCountry;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.Parameter;
import ac.ed.lurg.utils.LogWriter;

public abstract class AbstractDemandManager {

	protected CompositeCountryManager compositeCountryManager;
	protected CalorieManager calorieManager;
	protected BioenergyDemandManager bioenergyDemandManager;
	protected CerealFractionsManager cerealFractionsManager;

	public AbstractDemandManager(CompositeCountryManager compositeCountryManager, CalorieManager calorieManager) {
		this.compositeCountryManager = compositeCountryManager;
		this.calorieManager = calorieManager; 
		bioenergyDemandManager = new BioenergyDemandManager();
		cerealFractionsManager = new CerealFractionsManager(compositeCountryManager);
	}

	public Map<CommodityType, Double> getDemand(CompositeCountry cc, int year, Map<CommodityType, Double> prices) {

		if (!ModelConfig.CHANGE_DEMAND_YEAR)
			year = ModelConfig.BASE_YEAR;

		Map<CommodityType, Double> demandMap = new HashMap<CommodityType, Double>();
		Map<CommodityType, Double> foodDemandMap;

		for (SingleCountry c : compositeCountryManager.getAllForCompositeCountry(cc)) {

			foodDemandMap = getFoodDemand(c, year, prices);

			for (CommodityType commodity : CommodityType.getAllFoodItems()) {

				Double food = foodDemandMap.get(commodity);
				double bioenergy = bioenergyDemandManager.getFirstGenBioenergyDemand(c, year, commodity);
				// LogWriter.println(String.format("Country %s comm %s: %f",
				// country, commodity, bioenergy));
				
				double newDemand = ((food == null ? 0.0 : food.doubleValue()) + bioenergy);
				Double prevDemand = demandMap.get(commodity);

				if (prevDemand != null)
					newDemand += prevDemand.doubleValue(); // aggregate if
															// composite country

				demandMap.put(commodity, newDemand);
			}
		}
		
		return demandMap;
	}

	protected abstract Map<CommodityType, Double> getFoodDemand(SingleCountry c, int year, Map<CommodityType, Double> prices);
	
	public void updateGdpLossesFromShock(CompositeCountry cc, int year, double shockMagnitude) {}

	public double getSecondGenBioenergyDemand(Timestep timestep) {
		int year = ModelConfig.CHANGE_DEMAND_YEAR ? timestep.getYear() : ModelConfig.BASE_YEAR;
		double d = bioenergyDemandManager.getSecondGenBioenergyDemand(year);
		LogWriter.println("Global gen2 bioenergy demand in " + timestep + " is " + d);
		return d;
	}
	
	public double getFirstGenBioenergyDemand(CompositeCountry cc, int year, CommodityType commodity) {

		if (!ModelConfig.CHANGE_DEMAND_YEAR)
			year = ModelConfig.BASE_YEAR;

		double bioenergyDemand = 0.0;
		for (SingleCountry c : compositeCountryManager.getAllForCompositeCountry(cc)) {
			
				double newDemand = bioenergyDemandManager.getFirstGenBioenergyDemand(c, year, commodity);
				bioenergyDemand += newDemand; // aggregate if composite country
		}
		return bioenergyDemand;
	}
	
	public Map<CommodityType, Map<CropType, Double>> getBaseDemandFracts(CompositeCountry cc) {
		return cerealFractionsManager.getBaseDemandFracts(cc);
	}

	protected Map<CommodityType, Double> dietaryAdjustment(Map<CommodityType, Double> originalFoodDemandMap, int year, Map<CommodityType, Double> kcalPerT) {
		
		double monoDemandAdj = (!ModelConfig.SHOCKS_POSSIBLE) ? 0.0 : ModelConfig.getParameter(Parameter.MONOGASTRIC_CHANGE_ANNUAL_RATE, year);
		double rumDemandAdj = (!ModelConfig.SHOCKS_POSSIBLE) ? 0.0 : ModelConfig.getParameter(Parameter.RUMINANT_CHANGE_ANNUAL_RATE, year);
		
		Double monogastricDemand = originalFoodDemandMap.get(CommodityType.MONOGASTRICS);
		double monogastricEnergyShortfall = (monogastricDemand != null)
				? monogastricDemand * monoDemandAdj * kcalPerT.get(CommodityType.MONOGASTRICS) : 0.0;

		Double ruminantDemand = originalFoodDemandMap.get(CommodityType.RUMINANTS);
		double ruminantsEnergyShortfall = (ruminantDemand != null)
				? ruminantDemand * rumDemandAdj * kcalPerT.get(CommodityType.RUMINANTS) : 0.0;

		Map<CommodityType, Double> updatedFoodDemandMap = new HashMap<CommodityType, Double>();

		for (CommodityType commodity : CommodityType.getAllFoodItems()) {
			Double originalCommodityDemand = originalFoodDemandMap.get(commodity);

			if (originalCommodityDemand != null) {
				if (commodity.isAnimalProduct()) {
					double adj = commodity.getGamsName() == "ruminants" ? rumDemandAdj : monoDemandAdj;
					updatedFoodDemandMap.put(commodity,
							originalCommodityDemand*(1-adj));
				} else {
					updatedFoodDemandMap.put(commodity,
							originalCommodityDemand
									+ commodity.getMeatSubstitutionProportion() * ruminantsEnergyShortfall
											/ kcalPerT.get(commodity)
									+ commodity.getMeatSubstitutionProportion() * monogastricEnergyShortfall
											/ kcalPerT.get(commodity));

				}
			}

		}
		
//		LogWriter.println("mono demand adj " + monoDemandAdj +" rum demand adj " + rumDemandAdj + " year " + year + " years of change " + yearsOfChange + " mono shortfall " + monogastricEnergyShortfall + "ruminant shortfall " +  ruminantsEnergyShortfall);
//		LogWriter.println("original map " + originalFoodDemandMap);
//		LogWriter.println("updated map " + updatedFoodDemandMap);
		return updatedFoodDemandMap;
	}

}
