package ac.ed.lurg.yield;

import java.io.File;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.Timestep;
import ac.ed.lurg.shock.YieldShockManager;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.FertiliserRate;
import ac.ed.lurg.types.IrrigationRate;
import ac.ed.lurg.types.Parameter;
import ac.ed.lurg.types.YieldType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.utils.WatchForFile;
import ac.sac.raster.AbstractTabularRasterReader;
import ac.sac.raster.RasterHeaderDetails;
import ac.sac.raster.RasterKey;

public class LPJYieldResponseMapReader {	

	private RasterHeaderDetails rasterProj;
	private double techYieldAdj;
	private YieldShockManager yieldShockManager;

	public LPJYieldResponseMapReader(RasterHeaderDetails rasterProj) {
		this.rasterProj = rasterProj;
		yieldShockManager = new YieldShockManager(rasterProj);
	}

	public YieldRaster getRasterData(Timestep timestep) {
		String rootDir = timestep.getYearSubDir(ModelConfig.YIELD_DIR);
		
		// wait for data to be available
		long startTime = System.currentTimeMillis();
		WatchForFile fileWatcher = new WatchForFile(new File(rootDir + File.separator + "done"));
		boolean foundFile = fileWatcher.await(ModelConfig.LPJG_MONITOR_TIMEOUT_SEC);
		if (!foundFile) {
			LogWriter.printlnError("Not able to find marker file.  May have timed out.");
			throw new RuntimeException("Not able to find marker file.  May have timed out.");
		}
		LogWriter.println("Found marker file in " + (System.currentTimeMillis() - startTime) + " ms");

		// arithmetic adjustment of yield for tech change  
		int timeStepForTechChange = ModelConfig.IS_CALIBRATION_RUN ? ModelConfig.START_TIMESTEP : timestep.getTimestep();  // we don't want technology to change in calibration, but sometime we run tech change backwards
		int yearsOfTech = (timeStepForTechChange - ModelConfig.TECHNOLOGY_CHANGE_START_STEP) * ModelConfig.TIMESTEP_SIZE;
		
		double techChangeAnnualRate = (!ModelConfig.SHOCKS_POSSIBLE)? ModelConfig.TECHNOLOGY_CHANGE_ANNUAL_RATE: ModelConfig.getParameter(Parameter.TECHNOLOGY_CHANGE_ANNUAL_RATE, timestep.getYear());
		
		techYieldAdj = 1.0 +  yearsOfTech * techChangeAnnualRate;
		LogWriter.println("Yield adjustment for technology for " + timestep + " is " + techYieldAdj);
		
		YieldRaster yieldRaster = new YieldRaster(rasterProj);
		readNppFile(yieldRaster, rootDir+File.separator+ModelConfig.ANPP_FILENAME);
		readYieldFile(yieldRaster, rootDir+File.separator+ModelConfig.YIELD_FILENAME);
		
		yieldShockManager.setShocks(yieldRaster, timestep);

		return yieldRaster;
	}
	
	private void readNppFile(YieldRaster yieldRaster, String fileToRead) {  // Passed in YieldRaster is updated
		
		AbstractTabularRasterReader<YieldResponsesItem> nppReader = new AbstractTabularRasterReader<YieldResponsesItem>("\\s+", 10, yieldRaster) {
			protected void setData(RasterKey key, YieldResponsesItem item, Map<String, Double> rowValues) {
				double adjFactor = 10 * ModelConfig.PASTURE_HARVEST_FRACTION / 0.446;  // 10 for kg/m2 to t/ha, 0.446 carbon units / dry matter, from Osaki, M., Shinano, T., Tadano, T., 1992. Carbon-nitrogen interaction in field crop production. Soil Science and Plant Nutrition 38, 553-564.
								
				for (FertiliserRate fert : FertiliserRate.values()) {
					double pastureYield;
					if (ModelConfig.PASTURE_FERT_RESPONSE_FROM_LPJ)
						pastureYield = Math.max(getValueForCol(rowValues, LPJCropTypes.PASTURE_C3 + "N" + fert.getId()), getValueForCol(rowValues, LPJCropTypes.PASTURE_C4 + "N" + fert.getId())) * adjFactor;
					else
						pastureYield = Math.max(getValueForCol(rowValues, LPJCropTypes.PASTURE_C3.toString()), getValueForCol(rowValues, LPJCropTypes.PASTURE_C4.toString())) * adjFactor;

					item.setYield(YieldType.getYieldType(fert, IrrigationRate.NO_IRRIG), CropType.PASTURE, pastureYield);
					item.setYield(YieldType.getYieldType(fert, IrrigationRate.MAX_IRRIG), CropType.PASTURE, pastureYield);
				}
			}
		};
		
		nppReader.getRasterDataFromFile(fileToRead);
	}

	private void readYieldFile(YieldRaster yieldRaster, String fileToRead) {  	// Passed in YieldRaster is updated
		
		AbstractTabularRasterReader<YieldResponsesItem> yieldReader = new AbstractTabularRasterReader<YieldResponsesItem>("\\s+", 10, yieldRaster) {
			private double adjFactor = 10;  // 10 for kg/m2 to t/ha
			
			protected void setData(RasterKey key, YieldResponsesItem item, Map<String, Double> rowValues) {
									
				for (FertiliserRate fert : FertiliserRate.values()) {
					for (IrrigationRate irrig : IrrigationRate.values()) {
						YieldType yieldType = YieldType.getYieldType(fert, irrig);
						String fertIrrigString = yieldType.getFertIrrigString();
						
						/*
						 * WARNING:  if you change mapping below you also need to change LpjgOutputer and IrrigationMaxAmountReader to match
						 */
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.WHEAT, LPJCropTypes.C3_CEREALS, fertIrrigString, ModelConfig.CALIB_FACTOR_CEREAL_C3);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.MAIZE, LPJCropTypes.C4_CEREALS, fertIrrigString, ModelConfig.CALIB_FACTOR_CEREAL_C4);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.ENERGY_CROPS, LPJCropTypes.C4_CEREALS, fertIrrigString, ModelConfig.CALIB_FACTOR_MISCANTHUS);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.RICE, LPJCropTypes.RICE, fertIrrigString, ModelConfig.CALIB_FACTOR_RICE);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.OILCROPS, LPJCropTypes.OILCROPS, fertIrrigString, ModelConfig.CALIB_FACTOR_OILCROPS);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.PULSES, LPJCropTypes.PULSES, fertIrrigString, ModelConfig.CALIB_FACTOR_PULSES);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.STARCHY_ROOTS, LPJCropTypes.STARCHY_ROOTS, fertIrrigString, ModelConfig.CALIB_FACTOR_STARCHY_ROOTS);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.FRUITVEG, LPJCropTypes.FRUITVEG, fertIrrigString, ModelConfig.CALIB_FACTOR_FRUITVEG);
						getAndSetCalibratedYield(item, yieldType, rowValues, CropType.SUGAR, LPJCropTypes.SUGAR, fertIrrigString, ModelConfig.CALIB_FACTOR_SUGAR);
					}
				}
			}
			
			protected void getAndSetCalibratedYield(YieldResponsesItem item, YieldType yieldType, Map<String, Double> rowValues, 
					CropType plumcrop, LPJCropTypes lpjcrop, String fertIrrigString, double calibFactor) {
				
				double yieldValue = getValueForCol(rowValues, lpjcrop + fertIrrigString);
				double yieldAdj = yieldValue * calibFactor * techYieldAdj  * adjFactor;
				item.setYield(yieldType, plumcrop, yieldAdj);  // applying the calibration factor can't go in YieldResponsesItem.setYield as this is also called from elsewhere
			}
		};
		
		yieldReader.getRasterDataFromFile(fileToRead);
	}
}
